Virtualized machine-check architecture.

From: Philip Soltero <psoltero@cs.unm.edu>


---
 Kconfig                                |   16 -
 palacios/include/palacios/vm_guest.h   |    9 
 palacios/include/palacios/vmm_mcheck.h |  157 ++++++
 palacios/src/palacios/Makefile         |    1 
 palacios/src/palacios/vm_guest.c       |   10 
 palacios/src/palacios/vmm_config.c     |    9 
 palacios/src/palacios/vmm_mcheck.c     |  900 ++++++++++++++++++++++++++++++++
 utils/guest_creator/default.xml        |    6 
 8 files changed, 1103 insertions(+), 5 deletions(-)
 create mode 100644 palacios/include/palacios/vmm_mcheck.h
 create mode 100644 palacios/src/palacios/vmm_mcheck.c

diff --git a/Kconfig b/Kconfig
index 7348a5f..3c2dbb0 100644
--- a/Kconfig
+++ b/Kconfig
@@ -183,6 +183,13 @@ config DEBUG_LINUX_VIRTIO_VNET
         help
           Enable debugging for the VNET Virtio interface
 
+config MCHECK
+    depends on EXPERIMENTAL
+    bool "Enable machine-check architecture support in Palacios"
+    default n
+    help
+        Provides the ability to inject machine-check exceptions into a guest.
+
 endmenu
 
 
@@ -394,9 +401,12 @@ config DEBUG_MSR
     help
         This turns on debugging for MSR handling.
  
-
-
-
+config DEBUG_MCHECK
+    bool "Machine-check architecture"
+    default n
+    depends on EXPERIMENTAL && DEBUG_ON
+    help
+        Machine-check architecture debugging.
 
 endmenu
 
diff --git a/palacios/include/palacios/vm_guest.h b/palacios/include/palacios/vm_guest.h
index b435390..0a9bee3 100644
--- a/palacios/include/palacios/vm_guest.h
+++ b/palacios/include/palacios/vm_guest.h
@@ -47,6 +47,11 @@
 struct v3_sym_core_state;
 #endif
 
+#ifdef CONFIG_MCHECK
+#include <palacios/vmm_mcheck.h>
+struct v3_mcheck_state;
+#endif
+
 #include <palacios/vmm_config.h>
 
 struct v3_intr_state;
@@ -175,6 +180,10 @@ struct v3_vm_info {
     struct v3_telemetry_state telemetry;
 #endif
 
+#ifdef CONFIG_MCHECK
+    struct v3_mcheck_state * mcheck_state;
+#endif
+
     uint64_t yield_cycle_period;  
 
 
diff --git a/palacios/include/palacios/vmm_mcheck.h b/palacios/include/palacios/vmm_mcheck.h
new file mode 100644
index 0000000..5f8faaa
--- /dev/null
+++ b/palacios/include/palacios/vmm_mcheck.h
@@ -0,0 +1,157 @@
+/*
+ * This file is part of the Palacios Virtual Machine Monitor developed
+ * by the V3VEE Project with funding from the United States National
+ * Science Foundation and the Department of Energy.
+ *
+ * The V3VEE Project is a joint project between Northwestern University
+ * and the University of New Mexico.  You can find out more at
+ * http://www.v3vee.org
+ *
+ * Copyright (c) 2008, The V3VEE Project <http://www.v3vee.org>
+ * All rights reserved.
+ *
+ * This is free software.  You are permitted to use,
+ * redistribute, and modify it as specified in the file "V3VEE_LICENSE".
+ */
+
+/**
+ * @file Virtualized machine-check architecture.
+ * Provides definitions for various Machine Check Architecture
+ * (MCA) structures. Definitions found in BIOS and Kernel Developers Guide for AMD family 10h
+ * processors.
+ *
+ * @author <a HREF="mailto:psoltero@cs.unm.edu.us">Philip Soltero</a>
+ */
+
+#ifndef __VMM_MCHECK_H__
+#define __VMM_MCHECK_H__
+
+#ifdef __V3VEE__
+
+#include <palacios/vm_guest.h>
+#include <palacios/vmm_config.h>
+#include <palacios/vmm_msr.h>
+
+/**
+ * MCA status low and high registers, MC4_STAT, MSR0000_0411.
+ */
+struct mc4_stat_msr {
+     union {
+        uint64_t value;
+        struct {
+            union {
+                uint32_t value;
+                struct {
+                    uint_t syndrome : 8;
+                    uint_t reserved : 3;
+                    uint_t error_code_ext : 5;
+                    uint_t error_code : 16;
+                } __attribute__((packed));
+            } __attribute__((packed)) lo;
+
+            union {
+                uint32_t value;
+                struct {
+                    uint_t err_cpu : 4;
+                    uint_t ltd_link : 4;
+                    uint_t scrub : 1;
+                    uint_t sublink : 1;
+                    uint_t mca_stat_sub_cache : 2;
+                    uint_t reserved_01 : 1;
+                    uint_t uecc : 1;
+                    uint_t cecc : 1;
+                    uint_t syndrome : 8;
+                    uint_t reserved_02 : 1;
+                    uint_t err_cpu_val : 1;
+                    uint_t pcc : 1;
+                    uint_t addr_v : 1;
+                    uint_t misc_v : 1;
+                    uint_t en : 1;
+                    uint_t uc : 1;
+                    uint_t over : 1;
+                    uint_t val : 1;
+                } __attribute__((packed));
+            } __attribute__((packed)) hi;
+        }__attribute__((packed));
+     }__attribute__((packed));
+} __attribute__((packed));
+
+/**
+ * MCA address low and high registers, MC4_ADDR, MSR0000_0412.
+ */
+struct mc4_addr_msr {
+    union {
+        uint64_t value;
+
+        struct {
+            union {
+                uint32_t value;
+                struct {
+                    uint32_t err_addr : 31;
+                    uint32_t reserved : 1;
+                } __attribute__((packed));
+            } __attribute__((packed)) lo;
+
+            union {
+                uint32_t value;
+                struct {
+                    uint32_t err_addr : 16;
+                    uint32_t reserved : 16;
+                } __attribute__((packed));
+            } __attribute__((packed)) hi;
+        } __attribute__((packed));
+    } __attribute__((packed));
+} __attribute__((packed));
+
+/**
+ * Configures virtual machine-check architecture support.
+ *
+ * @param vm Virtual machine shared state.
+ * @param vm_cfg Virtual machine configuration tree.
+ *
+ * @return 0 for success and -1 for failure.
+ */
+int v3_config_mcheck(struct v3_vm_info * const vm,
+                     v3_cfg_tree_t * const vm_cfg);
+
+/**
+ * Deinitialize virtual machine-check architecture support.
+ *
+ * @param vm Virtual machine shared state.
+ */
+int v3_deinit_mcheck(struct v3_vm_info * const vm);
+
+/**
+ * Initialize the virtual machine-check architecture.
+ *
+ * @param vm Virtual machine shared state.
+ */
+int v3_init_mcheck(struct v3_vm_info * const vm);
+
+/**
+ * Injects a Northbridge machine-check exception on the specified cpu in the specified vm.
+ *
+ * @param vm Virtual machine shared state.
+ * @param cpu CPU to inject MCE on.
+ * @param stat MC4 status register.
+ * @param addr MC4 address register.
+ *
+ * @return 0 for sucess and -1 for failure.
+ */
+int v3_mcheck_inject_nb_mce(struct v3_vm_info * const vm,
+                            const uint32_t cpu,
+                            const struct mc4_stat_msr stat,
+                            const struct mc4_addr_msr addr);
+
+/**
+ * Determines if the MCA is available.
+ *
+ * @param vm Virtual machine shared state.
+ *
+ * @return 0 if the MCA is not available and 1 if it is.
+ */
+int v3_mcheck_mca_available(struct v3_vm_info * const vm);
+
+#endif /* __V3VEE__ */
+
+#endif /* __VMM_MCHECK_H__ */
diff --git a/palacios/src/palacios/Makefile b/palacios/src/palacios/Makefile
index f3b2c0c..bb28621 100644
--- a/palacios/src/palacios/Makefile
+++ b/palacios/src/palacios/Makefile
@@ -35,6 +35,7 @@ obj-y := \
 	vmm_muxer.o \
 	vmm_mem_hook.o \
 	vmm_mptable.o \
+	vmm_mcheck.o \
 
 obj-$(CONFIG_SVM) +=    svm.o \
 			svm_io.o \
diff --git a/palacios/src/palacios/vm_guest.c b/palacios/src/palacios/vm_guest.c
index c85c235..7abae01 100644
--- a/palacios/src/palacios/vm_guest.c
+++ b/palacios/src/palacios/vm_guest.c
@@ -28,6 +28,7 @@
 #include <palacios/vm_guest_mem.h>
 #include <palacios/vmm_lowlevel.h>
 #include <palacios/vmm_sprintf.h>
+#include <palacios/vmm_mcheck.h>
 #include <palacios/vmm_muxer.h>
 #include <palacios/vmm_xed.h>
 #include <palacios/vmm_direct_paging.h>
@@ -514,6 +515,9 @@ int v3_init_vm(struct v3_vm_info * vm) {
 
     v3_init_time_vm(vm);
 
+#ifdef CONFIG_MCHECK
+    v3_init_mcheck(vm);
+#endif
 
 #ifdef CONFIG_SYMBIOTIC
     v3_init_symbiotic_vm(vm);
@@ -556,12 +560,14 @@ int v3_free_vm_internal(struct v3_vm_info * vm) {
 
     v3_remove_hypercall(vm, GUEST_INFO_HCALL);
 
-
-
 #ifdef CONFIG_SYMBIOTIC
     v3_deinit_symbiotic_vm(vm);
 #endif
 
+#ifdef CONFIG_MCHECK
+    v3_deinit_mcheck(vm);
+#endif
+
     // init SVM/VMX
     switch (cpu_type) {
 #ifdef CONFIG_SVM
diff --git a/palacios/src/palacios/vmm_config.c b/palacios/src/palacios/vmm_config.c
index 50e3b1c..80d256c 100644
--- a/palacios/src/palacios/vmm_config.c
+++ b/palacios/src/palacios/vmm_config.c
@@ -32,6 +32,7 @@
 #include <palacios/vmm_msr.h>
 #include <palacios/vmm_mptable.h>
 #include <palacios/vmm_sprintf.h>
+#include <palacios/vmm_mcheck.h>
 
 
 
@@ -259,6 +260,7 @@ static int pre_config_vm(struct v3_vm_info * vm, v3_cfg_tree_t * vm_cfg) {
     }
 #endif
 
+
     if (v3_init_vm(vm) == -1) {
 	PrintError("Failed to initialize VM\n");
 	return -1;
@@ -368,6 +370,13 @@ static int post_config_vm(struct v3_vm_info * vm, v3_cfg_tree_t * cfg) {
 	return -1;
     }
     
+#ifdef CONFIG_MCHECK
+    if (v3_config_mcheck(vm, cfg) == -1) {
+        PrintError("Failed to configure machine-check architecture support.\n");
+        return -1;
+    }
+#endif
+
     //v3_hook_io_port(info, 1234, &IO_Read, NULL, info);
   
     if (setup_devices(vm, cfg) == -1) {
diff --git a/palacios/src/palacios/vmm_mcheck.c b/palacios/src/palacios/vmm_mcheck.c
new file mode 100644
index 0000000..5c41050
--- /dev/null
+++ b/palacios/src/palacios/vmm_mcheck.c
@@ -0,0 +1,900 @@
+/*
+ * This file is part of the Palacios Virtual Machine Monitor developed
+ * by the V3VEE Project with funding from the United States National
+ * Science Foundation and the Department of Energy.
+ *
+ * The V3VEE Project is a joint project between Northwestern University
+ * and the University of New Mexico.  You can find out more at
+ * http://www.v3vee.org
+ *
+ * Copyright (c) 2008, The V3VEE Project <http://www.v3vee.org>
+ * All rights reserved.
+ *
+ * This is free software.  You are permitted to use,
+ * redistribute, and modify it as specified in the file "V3VEE_LICENSE".
+ */
+
+/**
+ * @file Virtualized machine-check architecture.
+ *
+ * @author <a HREF="mailto:psoltero@cs.unm.edu.us">Philip Soltero</a>
+ */
+
+#include <palacios/vmm.h>
+#include <palacios/vmm_excp.h>
+#include <palacios/vmm_lowlevel.h>
+#include <palacios/vmm_mcheck.h>
+#include <palacios/vmm_msr.h>
+#include <palacios/vmm_string.h>
+
+#ifndef CONFIG_DEBUG_MCHECK
+#undef PrintDebug
+#define PrintDebug(fmt, args...)
+#endif
+
+#define CPUID_0000_0001 0x00000001
+#define CPUID_8000_0001 0x80000001
+
+// 6 error reporting banks. This may be configurable in the future.
+#define MCE_BANKS 6
+#define MCE_INTERRUPT 18
+
+#define MCA_AVAILABLE_BIT 0x00004000
+#define MCE_AVAILABLE_BIT 0x00000080
+#define MCA_AVAILABLE_TAG "mca_available"
+#define MCA_CTL_AVAILABLE_TAG "mca_ctl_available"
+#define MCA_YES "yes"
+#define MCA_NO "no"
+
+#define MSG_PRE "MCHECK: "
+
+#define MCG_CAP  0x0179
+#define MCG_STAT 0x017A
+#define MCG_CTL  0x017B
+
+#define MC0_CTL  0x0400
+#define MC0_STAT 0x0401
+#define MC0_ADDR 0x0402
+#define MC0_MISC 0x0403
+
+#define MC1_CTL  0x0404
+#define MC1_STAT 0x0405
+#define MC1_ADDR 0x0406
+#define MC1_MISC 0x0407
+
+#define MC2_CTL  0x0408
+#define MC2_STAT 0x0409
+#define MC2_ADDR 0x040A
+#define MC2_MISC 0x040B
+
+#define MC3_CTL  0x040C
+#define MC3_STAT 0x040D
+#define MC3_ADDR 0x040E
+#define MC3_MISC 0x040F
+
+#define MC4_CTL  0x0410
+#define MC4_STAT 0x0411
+#define MC4_ADDR 0x0412
+#define MC4_MISC 0x0413
+
+#define MC5_CTL  0x0414
+#define MC5_STAT 0x0415
+#define MC5_ADDR 0x0416
+#define MC5_MISC 0x0417
+
+/**
+ * Global machine-check capabilities register, MCG_CAP.
+ */
+struct mcg_cap_msr {
+    union {
+        uint64_t value;
+        struct {
+            uint32_t count : 8;
+            uint32_t mcg_ctl_p : 1;
+            uint64_t reserved : 55;
+        } __attribute__((packed));
+    } __attribute__((packed));
+} __attribute__((packed));
+
+/**
+ * Global machine-check status register, MCG_STAT.
+ */
+struct mcg_stat_msr {
+    union {
+        uint64_t value;
+        struct {
+            uint32_t ripv : 1;
+            uint32_t eipv : 1;
+            uint32_t mcip : 1; // Machine-check in progress.
+            uint64_t reserved : 61;
+        } __attribute__((packed));
+    } __attribute__((packed));
+} __attribute__((packed));
+
+/**
+ * Global machine-check control register, MCG_CTL.
+ */
+struct mcg_ctl_msr {
+    union {
+        uint64_t value;
+        struct {
+            uint32_t dce : 1; // Data cache register bank enable
+            uint32_t ice : 1; // Instruction cache register bank enable
+            uint32_t bue : 1; // Bus unit register bank enable
+            uint32_t lse : 1; // Load-store register bank enable
+            uint32_t nbe : 1; // Northbridge register bank enable
+            uint32_t fre : 1; // Fixed issue reorder buffer register bank enable
+            uint64_t unused : 58;
+        } __attribute__((packed));
+    } __attribute__((packed));
+} __attribute__((packed));
+
+/**
+ * A temporary structure for unimplemented machine-check error reporting banks.
+ */
+struct generic_bank {
+    struct v3_msr ctl;
+    struct v3_msr stat;
+    struct v3_msr addr;
+    struct v3_msr misc;
+} __attribute__((packed));
+
+struct v3_mcheck_state {
+    struct {
+        uint32_t mca_available : 1;      // Machine-check architecture configured available.
+        uint32_t mce_available : 1;      // Machine-check exception configured available.
+        uint32_t phys_mca_available : 1; // Physical machine-check architecture available.
+        uint32_t phys_mce_available : 1; // Physical machine-check exception available.
+    } __attribute__((packed)) status;
+    struct {
+        struct mcg_cap_msr cap;
+        struct mcg_stat_msr stat;
+        struct mcg_ctl_msr ctl;
+    } __attribute__((packed)) mcg;
+    struct generic_bank mc0;
+    struct generic_bank mc1;
+    struct generic_bank mc2;
+    struct generic_bank mc3;
+    struct {
+        uint64_t ctl;
+        struct mc4_stat_msr stat;
+        struct mc4_addr_msr addr;
+        uint64_t misc;
+    } __attribute__((packed)) mc4;
+    struct generic_bank mc5;
+} __attribute__((packed));
+
+static inline int virtualized_cpuids(struct v3_mcheck_state * const state);
+
+/**
+ * Handles a guest read of cpuid function 0000_0001 and 8000_0001.
+ * All bits are passthrough except for bit 14, the MCA available bit, and bit 7, the MCE available
+ * bit.
+ *
+ * @b<Note:> The virtual MCA only uses two bits in the entire 256 bit "return value". If other VMM
+ * subsystems or devices require the virtualization of other return value bits, it is suggested that
+ * this hook handler be moved to a common source file where all subsystems and devices can
+ * virtualize the bits they need to.
+ */
+static int cpuid_hook_handler(
+    struct guest_info * const info, const uint32_t cpuid, uint32_t * const eax,
+    uint32_t * const ebx, uint32_t * const ecx, uint32_t * const edx, void * const private_data) {
+    const struct v3_mcheck_state * const state = (struct v3_mcheck_state *)private_data;
+
+    // Most bits are passthrough.
+    v3_cpuid(cpuid, eax, ebx, ecx, edx);
+
+    // Bit 7, MCE availability, is virtualized.
+    if (state->status.mce_available) {
+        *edx |= 0x00000080;
+    }
+    else {
+        *edx &= 0xFFFFFF7F;
+    }
+
+    // Bit 14, MCA availability, is virtualized.
+    if (state->status.mca_available) {
+        *edx |= 0x00004000;
+    }
+    else {
+        *edx &= 0xFFFFBFFF;
+    }
+
+    return 0;
+}
+
+static inline
+void init_state(struct v3_mcheck_state * const state) {
+    memset(state, 0, sizeof(struct v3_mcheck_state));
+}
+
+/**
+ * Handles guest reads to MCA MSRs.
+ */
+static
+int msr_read_handler(struct guest_info * const core,
+                     const uint_t msr,
+                     struct v3_msr * const dst,
+                     void * const priv_data) {
+    struct v3_mcheck_state * const state = (struct v3_mcheck_state *)priv_data;
+    const uint8_t ctl_present = state->mcg.cap.mcg_ctl_p;
+
+    switch(msr) {
+    case MCG_CAP:
+        dst->value = state->mcg.cap.value;
+        break;
+
+    case MCG_STAT:
+        dst->value = state->mcg.stat.value;
+        break;
+
+    case MCG_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mcg.ctl.value;
+        break;
+
+    case MC0_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mc0.ctl.value;
+        break;
+
+    case MC0_STAT:
+        dst->value = state->mc0.stat.value;
+        break;
+
+    case MC0_ADDR:
+        dst->value = state->mc0.addr.value;
+        break;
+
+    case MC0_MISC:
+        dst->value = state->mc0.misc.value;
+        break;
+
+    case MC1_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mc1.ctl.value;
+        break;
+
+    case MC1_STAT:
+        dst->value = state->mc1.stat.value;
+        break;
+
+    case MC1_ADDR:
+        dst->value = state->mc1.addr.value;
+        break;
+
+    case MC1_MISC:
+        dst->value = state->mc1.misc.value;
+        break;
+
+    case MC2_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mc2.ctl.value;
+        break;
+
+    case MC2_STAT:
+        dst->value = state->mc2.stat.value;
+        break;
+
+    case MC2_ADDR:
+        dst->value = state->mc2.addr.value;
+        break;
+
+    case MC2_MISC:
+        dst->value = state->mc2.misc.value;
+        break;
+
+    case MC3_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mc3.ctl.value;
+        break;
+
+    case MC3_STAT:
+        dst->value = state->mc3.stat.value;
+        break;
+
+    case MC3_ADDR:
+        dst->value = state->mc3.addr.value;
+        break;
+
+    case MC3_MISC:
+        dst->value = state->mc3.misc.value;
+        break;
+
+    case MC4_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mc4.ctl;
+        break;
+
+    case MC4_STAT:
+        dst->value = state->mc4.stat.value;
+        break;
+
+    case MC4_ADDR:
+        dst->value = state->mc4.addr.value;
+        break;
+
+    case MC4_MISC:
+        dst->value = state->mc4.misc;
+        break;
+
+    case MC5_CTL:
+        chk_true(ctl_present, err_ctl_msr);
+        dst->value = state->mc5.ctl.value;
+        break;
+
+    case MC5_STAT:
+        dst->value = state->mc5.stat.value;
+        break;
+
+    case MC5_ADDR:
+        dst->value = state->mc5.addr.value;
+        break;
+
+    case MC5_MISC:
+        dst->value = state->mc5.misc.value;
+        break;
+
+    default:
+        PrintError(MSG_PRE "Ignoring read of unhooked MSR '0x%x'. This is a bug.\n", msr);
+    }
+
+    PrintDebug(MSG_PRE "Reading value '0x%llx' for MSR '0x%x'.\n", dst->value, msr);
+
+    return 0;
+
+    err_ctl_msr:
+    V3_Print(MSG_PRE "Ignoring read of control MSR '0x%x'. Control MSRs not supported.\n", msr);
+
+    return 0;
+}
+
+/**
+ * Handles guest writes to MCA MSRs.
+ */
+static
+int msr_write_handler(
+    struct guest_info * const core,
+    const uint_t msr,
+    const struct v3_msr src,
+    void * const priv_data) {
+    struct v3_mcheck_state * const state = (struct v3_mcheck_state *)priv_data;
+    const uint8_t ctl_present = state->mcg.cap.mcg_ctl_p;
+    uint32_t lower_6;
+    uint64_t upper_58;
+
+    switch(msr) {
+    case MCG_CAP:
+        goto err_mcg_cap;
+
+    case MCG_STAT:
+        state->mcg.stat.value = 0;
+        break;
+
+    case MCG_CTL:
+        chk_true(ctl_present, err_ctl);
+
+        // The upper 58 bits are unused and read-only.
+        lower_6 = src.value & 0x3f;
+        upper_58 = src.value - lower_6;
+        state->mcg.ctl.value = lower_6;
+        chk_true((upper_58 == 0), err_ctl_write);
+        break;
+
+    case MC0_CTL:
+        chk_true(ctl_present, err_ctl);
+        state->mc0.ctl.value = src.value;
+        break;
+
+    case MC0_STAT:
+        chk_true((src.value == 0), err_stat);
+        state->mc0.stat.value = 0;
+        break;
+
+    case MC0_ADDR:
+        state->mc0.addr.value = src.value;
+        break;
+
+    case MC0_MISC:
+        goto err_misc;
+
+    case MC1_CTL:
+        chk_true(ctl_present, err_ctl);
+        state->mc1.ctl.value = src.value;
+        break;
+
+    case MC1_STAT:
+        chk_true((src.value == 0), err_stat);
+        state->mc1.stat.value = 0;
+        break;
+
+    case MC1_ADDR:
+        state->mc1.addr.value = src.value;
+        break;
+
+    case MC1_MISC:
+        goto err_misc;
+
+    case MC2_CTL:
+        chk_true(ctl_present, err_ctl);
+        state->mc2.ctl.value = src.value;
+        break;
+
+    case MC2_STAT:
+        chk_true((src.value == 0), err_stat);
+        state->mc2.stat.value = 0;
+        break;
+
+    case MC2_ADDR:
+        state->mc2.addr.value = src.value;
+        break;
+
+    case MC2_MISC:
+        goto err_misc;
+
+    case MC3_CTL:
+        chk_true(ctl_present, err_ctl);
+        state->mc3.ctl.value = src.value;
+        break;
+
+    case MC3_STAT:
+        chk_true((src.value == 0), err_stat);
+        state->mc3.stat.value = 0;
+        break;
+
+    case MC3_ADDR:
+        state->mc3.addr.value = src.value;
+        break;
+
+    case MC3_MISC:
+        goto err_misc;
+
+    case MC4_CTL:
+        chk_true(ctl_present, err_ctl);
+        state->mc4.ctl = src.value;
+        break;
+
+    case MC4_STAT:
+        chk_true((src.value == 0), err_stat);
+        state->mc4.stat.value = 0;
+        break;
+
+    case MC4_ADDR:
+        state->mc4.addr.value = src.value;
+        break;
+
+    case MC4_MISC:
+        goto err_misc;
+
+    case MC5_CTL:
+        chk_true(ctl_present, err_ctl);
+        state->mc5.ctl.value = src.value;
+        break;
+
+    case MC5_STAT:
+        chk_true((src.value == 0), err_stat);
+        state->mc5.stat.value = src.value;
+        break;
+
+    case MC5_ADDR:
+        state->mc5.addr.value = src.value;
+        break;
+
+    case MC5_MISC:
+        goto err_misc;
+
+    default:
+        PrintError(MSG_PRE "Ignoring write of unhooked MSR '0x%x'. This is a bug.\n", msr);
+    }
+
+    PrintDebug(MSG_PRE "Writing value '0x%llx' for MSR '0x%x'.\n", src.value, msr);
+
+    return 0;
+
+    err_mcg_cap:
+    V3_Print(MSG_PRE "Ignoring write to MCG_CAP MSR.\n");
+    return 0;
+
+    err_ctl:
+    V3_Print(MSG_PRE "Ignoring write to control MSR '0x%x'. Control MSRs not supported.\n", msr);
+    return 0;
+
+    err_ctl_write:
+    V3_Print(MSG_PRE "Ignoring write of '0x%llx' to MCG_CTL read-only bits.\n", upper_58);
+    return 0;
+
+    err_misc:
+    V3_Print(MSG_PRE "Ignoring write to read only miscellaneous MSR '0x%x'.\n", msr);
+    return 0;
+
+    err_stat:
+    // Should be a GPF.
+    V3_Print(MSG_PRE "Ignoring write of illegal value '0x%llx'.\n", src.value);
+    return 0;
+}
+
+/**
+ * Checks if the physical MCA and MCE are available.
+ */
+static inline
+void check_phys_availability(struct v3_mcheck_state * const state) {
+    uint32_t target = 0x00000001;
+    uint32_t eax;
+    uint32_t ebx;
+    uint32_t ecx;
+    uint32_t edx;
+
+    v3_cpuid(target, &eax, &ebx, &ecx, &edx);
+
+    state->status.phys_mca_available = (edx & MCA_AVAILABLE_BIT) ? 1 : 0;
+    state->status.phys_mce_available = (edx & MCE_AVAILABLE_BIT) ? 1 : 0;
+
+    PrintDebug(MSG_PRE "Physical MCA %s available.\n"
+               MSG_PRE "Physical MCE %s available.\n",
+               (state->status.phys_mca_available) ? "is" : "is not",
+               (state->status.phys_mce_available) ? "is" : "is not");
+}
+
+/**
+ * Configures MCA availability.
+ */
+static inline
+int config_mca_availability(
+    v3_cfg_tree_t * const vm_cfg,
+    struct v3_mcheck_state * const state) {
+    char * mca_enable_str;
+
+    // Either the user specified an 'mca_available' tag or did not.
+    // If so, the 'mca_available' value will be either 'yes' or 'no'.
+    // If not, default to 'yes' to avoid confusion and print a warning.
+
+    mca_enable_str = v3_cfg_val(vm_cfg, MCA_AVAILABLE_TAG);
+    if (strcasecmp(mca_enable_str, MCA_YES) == 0) {
+        state->status.mca_available = 1;
+        PrintDebug(MSG_PRE "MCA is configured as available.\n");
+    }
+    else if (strcasecmp(mca_enable_str, MCA_NO) == 0) {
+        state->status.mca_available = 0;
+        PrintDebug(MSG_PRE "MCA is configured as not available.\n");
+    }
+    else if (strcasecmp(mca_enable_str, "") == 0) {
+        state->status.mca_available = 1;
+        PrintDebug(MSG_PRE "No mca_available configuration tag specified. "
+                  "Defaulting to mca available.\n");
+    }
+    else {
+        PrintError(MSG_PRE "Unknown value '%s' for mca_available tag.\n", mca_enable_str);
+        return -1;
+    }
+
+    return 0;
+}
+
+/**
+ * Configures MCA control MSR availability.
+ */
+static inline
+int config_mca_ctl_availability(
+    v3_cfg_tree_t * const vm_cfg,
+    struct v3_mcheck_state * const state) {
+    char * const mca_ctl_available_str = v3_cfg_val(vm_cfg, MCA_CTL_AVAILABLE_TAG);
+
+    // Either the user specified an 'mca_ctl_available' tag or did not.
+    // If so, the 'mca_ctl_available' value will be either 'yes' or 'no'.
+    // If not, default to 'yes' to avoid confusion and print a warning.
+
+    if (strcasecmp(mca_ctl_available_str, MCA_YES) == 0) {
+        state->mcg.cap.mcg_ctl_p = 1;
+        PrintDebug(MSG_PRE "CTL MSRs configured as available.\n");
+    }
+    else if (strcasecmp(mca_ctl_available_str, MCA_NO) == 0) {
+        state->mcg.cap.mcg_ctl_p = 0;
+        PrintDebug(MSG_PRE "CTL MSRs configured as not available.\n");
+    }
+    else if (strcasecmp(mca_ctl_available_str, "") == 0) {
+        state->mcg.cap.mcg_ctl_p = 1;
+        PrintDebug(MSG_PRE "No mca_ctl_available configuration tag specified. "
+                  "Defaulting to ctl available.\n");
+    }
+    else {
+        PrintError(MSG_PRE "Unknown value '%s' for mca_ctl_available tag.\n",
+                   mca_ctl_available_str);
+        return -1;
+    }
+
+    return 0;
+}
+
+/**
+ * Configures machine-check exception availability.
+ */
+static inline
+int config_mce_availability(
+    v3_cfg_tree_t * const vm_cfg,
+    struct v3_mcheck_state * const state) {
+    // I'm not sure what the implications of making the MCE available without the MCA and vice-versa
+    // are, so for now if the MCA is available then so is the MCE.
+    state->status.mce_available = state->status.mca_available;
+    return 0;
+}
+
+/**
+ * Hooks CPUID functions if necessary.
+ * CPUID functions 0000_0001 and 8000_0001 are hooked if the following conditions are true:
+ *  - The physical MCA is available but the virtual MCA is not configured as available
+ *  - The physical MCA is not available but the virtual MCA is configured as available
+ *
+ * @param phys_mca_available A flag indicating that the physical MCA is available.
+ * @param virt_mca_availble A flag indicating that the virtual MCA is configured as available.
+ *
+ * @return 0 for success and -1 for failure.
+ */
+static inline
+int hook_cpuids(
+    struct v3_vm_info * const vm,
+    struct v3_mcheck_state * const state) {
+    int ret;
+
+    chk_true((virtualized_cpuids(state) == 1), done);
+
+    ret = v3_hook_cpuid(vm, CPUID_0000_0001, cpuid_hook_handler, state);
+    chk_true((ret == 0), err_0000_0001);
+
+    ret = v3_hook_cpuid(vm, CPUID_8000_0001, cpuid_hook_handler, state);
+    chk_true((ret == 0), err_8000_0001);
+
+    done:
+    return 0;
+
+    err_8000_0001:
+    PrintError(MSG_PRE "Failed to hook CPUID function 8000_0001.\n");
+    goto fail_8000_0001;
+
+    err_0000_0001:
+    PrintError(MSG_PRE "Failed to hook CPUID function 0000_0001.\n");
+    goto fail_0000_0001;
+
+    fail_8000_0001:
+    v3_unhook_cpuid(vm, CPUID_0000_0001);
+
+    fail_0000_0001:
+    return -1;
+}
+
+/**
+ * Hook the MCA MSRs: MCG_CAP, MCG_STAT, and MCi_*.
+ */
+static inline
+int hook_msrs(
+    struct v3_vm_info * const vm,
+    struct v3_mcheck_state * const state) {
+    int ret;
+    int msr;
+
+    ret = v3_hook_msr(vm, MCG_CAP, msr_read_handler, msr_write_handler, state);
+    chk_true((ret == 0), err_cap);
+
+    ret = v3_hook_msr(vm, MCG_STAT, msr_read_handler, msr_write_handler, state);
+    chk_true((ret == 0), err_stat);
+
+    ret = v3_hook_msr(vm, MCG_CTL, msr_read_handler, msr_write_handler, state);
+    chk_true((ret == 0), err_ctl);
+
+    for (msr = MC0_CTL; msr <= MC5_MISC; ++msr) {
+         ret = v3_hook_msr(vm, msr, msr_read_handler, msr_write_handler, state);
+         chk_true((ret == 0), err_mci);
+     }
+
+    return 0;
+
+    err_mci:
+    PrintError(MSG_PRE "Failed to hook MCi MSR '0x%x'.\n", msr);
+    goto fail_mci;
+
+    err_ctl:
+    PrintError(MSG_PRE "Failed to hook MCG_CTL MSR.\n");
+    goto fail_ctl;
+
+    err_stat:
+    PrintError(MSG_PRE "Failed to hook MCG_STAT MSR.\n");
+    goto fail_stat;
+
+    err_cap:
+    PrintError(MSG_PRE "Failed to hook MCG_CAP MSR.\n");
+    goto fail_cap;
+
+    fail_mci:
+    v3_unhook_msr(vm, MCG_CTL);
+    for (msr = msr - 1; msr >= MC0_CTL; ++msr) {
+        v3_unhook_msr(vm, msr);
+    }
+
+    fail_ctl:
+    v3_unhook_msr(vm, MCG_STAT);
+
+    fail_stat:
+    v3_unhook_msr(vm, MCG_CAP);
+
+    fail_cap:
+    return -1;
+}
+
+static inline
+void unhook_cpuids(
+    struct v3_vm_info * const vm,
+    struct v3_mcheck_state * const state ) {
+
+    chk_true((virtualized_cpuids(state) == 1), done);
+
+    v3_unhook_cpuid(vm, CPUID_8000_0001);
+    v3_unhook_cpuid(vm, CPUID_0000_0001);
+
+    done:
+    return;
+}
+
+static inline
+void unhook_msrs(struct v3_vm_info * const vm) {
+    int msr;
+
+    for (msr = MC0_CTL; msr <= MC5_MISC; ++msr) {
+        v3_unhook_msr(vm, msr);
+    }
+
+    v3_unhook_msr(vm, MCG_CTL);
+    v3_unhook_msr(vm, MCG_STAT);
+    v3_unhook_msr(vm, MCG_CAP);
+}
+
+int v3_config_mcheck(struct v3_vm_info * const vm, v3_cfg_tree_t * const vm_cfg) {
+    v3_cfg_tree_t * mcheck_cfg;
+    struct v3_mcheck_state * state = vm->mcheck_state;
+    int ret;
+
+    mcheck_cfg = v3_cfg_subtree(vm_cfg, "mcheck");
+    chk_true((mcheck_cfg != NULL), done);
+
+    V3_Print(MSG_PRE "Virtual machine-check architecture support.\n");
+
+    ret = config_mca_availability(mcheck_cfg, state);
+    chk_true((ret == 0), fail_config_mcheck);
+
+    ret = config_mce_availability(mcheck_cfg, state);
+    chk_true((ret == 0), fail_config_mcheck);
+
+    ret = config_mca_ctl_availability(mcheck_cfg, state);
+    chk_true((ret == 0), fail_config_mcheck);
+
+    done:
+    return 0;
+
+    fail_config_mcheck:
+    return -1;
+}
+
+int v3_deinit_mcheck(struct v3_vm_info * const vm) {
+    struct v3_mcheck_state * state = vm->mcheck_state;
+    chk_true((state != NULL), done);
+
+    unhook_msrs(vm);
+    unhook_cpuids(vm, state);
+    V3_Free(state);
+    vm->mcheck_state = NULL;
+
+    done:
+    return 0;
+}
+
+int v3_init_mcheck(struct v3_vm_info * const vm) {
+    struct v3_mcheck_state * state;
+    int ret;
+
+    state = (struct v3_mcheck_state *)V3_Malloc(sizeof(struct v3_mcheck_state));
+    chk_true((state != NULL), err_state_malloc);
+
+    init_state(state);
+
+    vm->mcheck_state = state;
+
+    check_phys_availability(state);
+
+    state->mcg.cap.count = MCE_BANKS;
+
+    ret = hook_cpuids(vm, state);
+    chk_true((ret == 0), err_hook_cpuids);
+
+    ret = hook_msrs(vm, state);
+    chk_true((ret == 0), err_hook_msrs);
+
+    return 0;
+
+    err_hook_msrs:
+    PrintError(MSG_PRE "Error hooking MSRs.\n");
+    goto fail_hook_msrs;
+
+    err_hook_cpuids:
+    PrintError(MSG_PRE "Error hooking CPUIDs.\n");
+    goto fail_hook_cpuids;
+
+    err_state_malloc:
+    PrintError(MSG_PRE "Failed to allocate machine-check architecture state.\n");
+    goto fail_init_mcheck;
+
+    fail_hook_msrs:
+    unhook_cpuids(vm, state);
+
+    fail_hook_cpuids:
+    V3_Free(state);
+
+    fail_init_mcheck:
+    return -1;
+}
+
+int v3_mcheck_mca_available(struct v3_vm_info * const vm) {
+    return ((struct v3_mcheck_state *)vm->mcheck_state)->status.mca_available;
+}
+
+int v3_mcheck_inject_nb_mce(
+    struct v3_vm_info * const vm,
+    const uint32_t cpu,
+    const struct mc4_stat_msr stat,
+    const struct mc4_addr_msr addr) {
+    struct v3_mcheck_state * const state = vm->mcheck_state;
+    int ret;
+
+    // For now only MCE injection on cpu 0 is supported.
+    chk_true((cpu == 0), err_cpu);
+
+    // Is virtual MCA support enabled?
+    chk_true((state->status.mca_available == 1), err_mca);
+
+    // Is the Northbridge bank enabled?
+    chk_true((state->mcg.ctl.nbe == 1), err_nbe);
+
+    state->mc4.stat = stat;
+    state->mc4.addr = addr;
+
+    state->mcg.stat.value = 0;
+    state->mcg.stat.ripv = 1;
+    state->mcg.stat.mcip = 1;
+
+    PrintDebug(MSG_PRE "Injecting NB MCE on core %u.\n", 0);
+
+    // Raise on core 0.
+    ret = v3_raise_exception(&vm->cores[0], MCE_INTERRUPT);
+    chk_true((ret == 0), err_v3_raise_excp);
+
+    return 0;
+
+    err_v3_raise_excp:
+    PrintError(MSG_PRE "Failed to raise MCE.\n");
+    return -1;
+
+    err_nbe:
+    V3_Print(MSG_PRE "Northbridge register bank disabled. Ignoring Northbridge MCE.\n");
+    return 0;
+
+    err_mca:
+    PrintError(MSG_PRE "Virtual MCA support not enabled.\n");
+    return -1;
+
+    err_cpu:
+    PrintError(MSG_PRE "Injecting MCE on cpu %u not supported.\n", cpu);
+    return -1;
+}
+
+/**
+ * Determines if machine-check architecture related CPUIDs are virtualized.
+ */
+static inline
+int virtualized_cpuids(struct v3_mcheck_state * const state) {
+    if (state->status.phys_mca_available && state->status.mca_available) {
+        return 0;
+    }
+
+    if (!state->status.phys_mca_available && !state->status.mca_available) {
+        return 0;
+    }
+
+    return 1;
+}
diff --git a/utils/guest_creator/default.xml b/utils/guest_creator/default.xml
index fe5e67b..9efd032 100644
--- a/utils/guest_creator/default.xml
+++ b/utils/guest_creator/default.xml
@@ -20,6 +20,12 @@
 
 	<cores count="1"><core /></cores>
 
+        <!-- Machine-check architecture
+        <mcheck>
+                <mca_available>yes</mca_available>
+                <mca_ctl_available>yes</mca_ctl_available>
+        </mcheck>
+        -->
 
 	<!-- Memory redirections -->
 	<!-- Redirects a guest's memory region to a host physical address -->
